/*
 * C6Run_mem.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Memory allocation routines for C6Run buffers for remote preocedures  *
************************************************************************/


/*  ----------------------------------- Standard Headers              */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

/*  ----------------------------------- Linux Posix Headers           */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>

/*  ----------------------------------- CMEM API                      */
#include <ti/sdo/linuxutils/cmem/include/cmem.h>

/*  ----------------------------------- DSP/BIOS Link                 */
#include <dsplink.h>
#include <errbase.h>

/*  ----------------------------------- Application headers           */
#include "debug_common.h"
#include "ipc_common.h"
#include "C6Run.h"
#include "dlmalloc.h"

/*  ----------------------------------- This module's header          */
#include "C6Run_mem.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

#define USE_DLMALLOC_DIRECT

#define C6RUN_MEM_MALLOC_MAGIC        (0x43365231)
#define C6RUN_MEM_MEMALIGN_MAGIC      (0x43365232)

#define SET_MALLOC_MAGIC(P)           (C6RUN_MEM_MALLOC_MAGIC ^ ((uint32_t)P))
#define IS_MALLOC_MAGIC(M,P)          ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MALLOC_MAGIC)
#define SET_MEMALIGN_MAGIC(P)         (C6RUN_MEM_MEMALIGN_MAGIC ^ ((uint32_t)P))
#define IS_MEMALIGN_MAGIC(M,P)        ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MEMALIGN_MAGIC)

#define MAX_SIZE_T                    (~(size_t)0)

#define NUM_TABLE_BINS_BITS           (12)
#define NUM_TABLE_BINS                (0x1 << NUM_TABLE_BINS_BITS)

#define MIN_ALIGNMENT_BITS            (3)
#define MIN_ALIGNMENT                 (0x1 << MIN_ALIGNMENT_BITS)

#define GET_BIN_NUMBER(P)             (((((uint32_t) P) - ((uint32_t)C6RUN_MEM_virtBaseAddr)) >> MIN_ALIGNMENT_BITS) & (NUM_TABLE_BINS - 1))

#define C6RUN_MEM_ALIGN(S)            ((((uint32_t)S) + MIN_ALIGNMENT) & (~(MIN_ALIGNMENT - 1)))

#define BUF2PTR(B) ((void *)((C6RUN_MEM_Buffer *)B + 1))
#define PTR2BUF(P) ((pC6RUN_MEM_Buffer)((C6RUN_MEM_Buffer *)P - 1))


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

static void LOCAL_insertInBufferTable(void *ptr);
static void LOCAL_removeFromBufferTable(void *ptr);


/************************************************************
* Local Variable Definitions                                *
************************************************************/

static CMEM_AllocParams cmem_params = {CMEM_HEAP, CMEM_CACHED, MIN_ALIGNMENT};
static pthread_mutex_t buffTableMutex;


/************************************************************
* Global Variable Definitions                               *
************************************************************/

volatile bool  C6RUN_MEM_initialized    = FALSE;
void          *C6RUN_MEM_virtBaseAddr   = NULL;
void          *C6RUN_MEM_physBaseAddr   = NULL;

pC6RUN_MEM_Buffer C6RUN_MEM_bufferTable[NUM_TABLE_BINS];


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Memory subsystem initialization routines
int C6RUN_MEM_init  ( void )
{
  if (C6RUN_MEM_initialized)
  {
    return C6RUN_SUCCESS;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_init()\n");

  // Initialize CMEM
  if (CMEM_init() == -1)
  {  
    printf("CMEM_init() failed!\n");
    return C6RUN_FAIL;
  }
  VERBOSE_PRINT0("CMEM_init() passed!\n");

  // Allocate the entire CMEM region and DSP region.
  // We can perform user-space management in dlmalloc for the CMEM region
  // We can use items from the DSP code region (DSP heap and stack) by doing
  //   an address translation.
  C6RUN_MEM_virtBaseAddr = CMEM_alloc(DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE, &cmem_params);
  if (C6RUN_MEM_virtBaseAddr == NULL)
  {
    printf("CMEM_alloc() failed!\n");
    return C6RUN_FAIL;
  }
  C6RUN_MEM_physBaseAddr = (void *)CMEM_getPhys(C6RUN_MEM_virtBaseAddr);
  VERBOSE_PRINT0("CMEM_alloc() passed!\n");
  
  // Init the C6RUN_MEM_bufferTable to NULL values
  memset(C6RUN_MEM_bufferTable,0x00,(NUM_TABLE_BINS*sizeof(pC6RUN_MEM_Buffer)));
  
  // Init the mutex protecting the bufferTable
  pthread_mutex_init(&buffTableMutex,NULL);

  C6RUN_MEM_initialized = true;

  // Register memory exit routine for calling upon program termination
  atexit(C6RUN_MEM_exit);
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_init()\n");
  
  return C6RUN_SUCCESS;
}

void C6RUN_MEM_exit ( void )
{
  if (!C6RUN_MEM_initialized)
  {
    return;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_exit()\n");

  // Check for orphaned memory(leaks) and print it out
  for (int i=0; i<NUM_TABLE_BINS; i++)
  {
    pC6RUN_MEM_Buffer pBuffer = C6RUN_MEM_bufferTable[i];
    if (pBuffer != NULL)
    {
      do
      {  
        VERBOSE_PRINT2("Orphaned memory at bin %d, 0x%08X.\n",i,(uint32_t)BUF2PTR(pBuffer));
        pBuffer = pBuffer->next;
      } while (pBuffer != NULL);
    }
  }
  
  // Free up the entire CMEM region
  CMEM_free(C6RUN_MEM_virtBaseAddr,&cmem_params);
  
  if (CMEM_exit() < 0)
  {
    printf("CMEM_exit() failed!\n");
  }
  VERBOSE_PRINT0("CMEM_exit() passed!\n");

  C6RUN_MEM_initialized = false;
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_exit()\n");
}


// C6Run memory buffer management routines
void *C6RUN_MEM_malloc(size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  
  VERBOSE_PRINT1("C6RUN_MEM_malloc(%d)\n",size);
  
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }  

  // Calculate total size (align size to minimum alignment)
  totalSize = (size + sizeof(C6RUN_MEM_Buffer));
  totalSize = C6RUN_MEM_ALIGN(totalSize);
  
  // Call dlmalloc
  tmpPtr = dlmalloc(totalSize);
  
  if (tmpPtr)
  {
    pC6RUN_MEM_Buffer pBuffer;
    
    pthread_mutex_lock(&buffTableMutex);
    
    pBuffer = (pC6RUN_MEM_Buffer) tmpPtr;

    pBuffer->next = NULL;
    tmpPtr = BUF2PTR(tmpPtr);
    pBuffer->magic = SET_MALLOC_MAGIC(tmpPtr);
    
    // Insert the pBuffer into the buffer table
    LOCAL_insertInBufferTable(tmpPtr);
    
    pthread_mutex_unlock(&buffTableMutex);
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_malloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_realloc(void *oldPtr,size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  pC6RUN_MEM_Buffer pBuffer;
  
  VERBOSE_PRINT2("C6RUN_MEM_realloc(0x%08X,%d)\n",(uint32_t)oldPtr,size);
  
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (oldPtr == NULL)
  {
    return C6RUN_MEM_malloc(size);
  }
  
  // Calculate total size
  totalSize = (size + sizeof(C6RUN_MEM_Buffer));
  totalSize = C6RUN_MEM_ALIGN(totalSize);
  pBuffer = PTR2BUF(oldPtr);
  
  if (pBuffer && ! IS_MALLOC_MAGIC(pBuffer->magic,oldPtr))
  {
    // This could be a memaligned block
    if ( IS_MEMALIGN_MAGIC((((uint32_t *)oldPtr)[-1]),oldPtr) )
    {
      oldPtr = ((void**)oldPtr)[-2];
      pBuffer =  PTR2BUF(oldPtr);
    }  
    return NULL;
  }
  
  if (IS_MALLOC_MAGIC(pBuffer->magic,oldPtr))
  {
    tmpPtr = dlrealloc((void *)pBuffer,totalSize);
  }
  else
  {
    tmpPtr = NULL;
  }
  
  if (tmpPtr)
  {
    pBuffer = (pC6RUN_MEM_Buffer) tmpPtr;
    tmpPtr = BUF2PTR(pBuffer);
    // If the new pointer is not the same as the old one, we have
    // some updates we need to do in the buffer table
    if (tmpPtr != oldPtr)
    {
      pthread_mutex_lock(&buffTableMutex); 
    
      // Remove old pointer from table
      LOCAL_removeFromBufferTable(oldPtr);
      
      pBuffer->next = NULL;
      pBuffer->magic = SET_MALLOC_MAGIC(tmpPtr);
      
      // Add new pointer/buffer to table
      LOCAL_insertInBufferTable(tmpPtr);
      
      pthread_mutex_unlock(&buffTableMutex);
    }
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_realloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_memalign(size_t alignment, size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_memalign(%d,%d)\n",alignment,size);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  // We can just use malloc
  if (alignment <= MIN_ALIGNMENT)
  {
    return C6RUN_MEM_malloc(size);
  }
  
  // Check for non power-of-two alignment
  if ((alignment & (alignment - 1)))
  {
    return NULL;
  }
  
  // Aligment is at least MIN_ALIGNMENT<<1 bytes
  totalSize = (alignment-MIN_ALIGNMENT) + size;
  tmpPtr = C6RUN_MEM_malloc(totalSize);
  if (tmpPtr != NULL)
  {
    intptr_t ptr = (intptr_t)tmpPtr;
    if ((ptr % alignment) == 0)
    {
      return tmpPtr;
    }

    // Align the pointer
    ptr += ((-ptr) % alignment);
    
    // There is always enough space for the base pointer and the memalign magic word
    ((uint32_t*)ptr)[-1] = SET_MEMALIGN_MAGIC(ptr);
    ((void**)ptr)[-2] = tmpPtr;

    tmpPtr = (void*)ptr;
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_memalign returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_calloc(size_t nelem, size_t elsize)
{
  size_t totalSize;
  void*  tmpPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_calloc(%d,%d)\n",nelem,elsize);
  
  totalSize = nelem * elsize;
  tmpPtr  = C6RUN_MEM_malloc(totalSize);
  if (tmpPtr != NULL)
  {
    memset(tmpPtr, 0x00, totalSize);
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_calloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void C6RUN_MEM_free(void* ptr)
{
  VERBOSE_PRINT1("C6RUN_MEM_free(0x%08X)\n",(uint32_t)ptr);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (ptr != NULL)
  {
    pC6RUN_MEM_Buffer pBuffer;
    pthread_mutex_lock(&buffTableMutex); 
    pBuffer = PTR2BUF(ptr);
  
    if (! IS_MALLOC_MAGIC(pBuffer->magic,ptr))
    {
      // This could be a memaligned block
      if ( IS_MEMALIGN_MAGIC((((uint32_t *)ptr)[-1]),ptr) )
      {
        ptr = ((void**)ptr)[-2];
        pBuffer = PTR2BUF(ptr);
      }
    }
  
    if (IS_MALLOC_MAGIC(pBuffer->magic,ptr))
    {
      LOCAL_removeFromBufferTable(ptr);
      dlfree((void *)pBuffer);
    }

    pthread_mutex_unlock(&buffTableMutex); 
  }
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_free.\n");
}


// Memory cache management routines
void C6RUN_MEM_globalInv()
{
  CMEM_cacheInv(C6RUN_MEM_virtBaseAddr, DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE);
}

void C6RUN_MEM_globalWb()
{
  CMEM_cacheWb(C6RUN_MEM_virtBaseAddr, DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE);
}

void C6RUN_MEM_inv(void *ptr,size_t size)
{
  CMEM_cacheInv(ptr, size);
}

void C6RUN_MEM_wb(void *ptr,size_t size)
{
  CMEM_cacheWb(ptr, size);
}


// Memory helper routines
void *C6RUN_MEM_lookupBuffer(void *ptr, size_t *buffSize)
{
  uint32_t binNum = GET_BIN_NUMBER(ptr);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];
  uint32_t base, size;
  
  // We check, but this should never happen
  if (pTmpBuffer)
  {
    // Search through buffer table bin for a match. A match is the
    // input pointer inside the range of an allocated buffer
    base = (uint32_t) BUF2PTR(pTmpBuffer);
    // Get size from dlmalloc header data (chunk size minus dlmalloc 
    // header size minus C6RUN_MEM_Buffer size)
    size = (uint32_t) (((size_t*)pTmpBuffer)[-1] & 0xFFFFFFFC) - (2 * sizeof(size_t)) - sizeof(C6RUN_MEM_Buffer);
    
    if ((base >= ((uint32_t) ptr)) && (((uint32_t) ptr) < (base + size)))
    {
      *buffSize = size;
      return ((void *)base);
    }
    else
    {
      while (pTmpBuffer->next != NULL)
      {
        base = (uint32_t) BUF2PTR(pTmpBuffer->next);
        size = (uint32_t) (((size_t*)pTmpBuffer->next)[-1] & 0xFFFFFFFC) - (2 * sizeof(size_t)) - sizeof(C6RUN_MEM_Buffer);
        if ((base >= ((uint32_t) ptr)) && (((uint32_t) ptr) < (base + size)))
        {
          *buffSize = size;
          return ((void *)base);
        }
        else
        {
          pTmpBuffer = pTmpBuffer->next;
        }
      }
    }
  }
  
  *buffSize = 0;
  return NULL;
}

void *C6RUN_MEM_virt2Phys(void *virtPtr)
{
  return (void *) (((virtPtr) - C6RUN_MEM_virtBaseAddr) + C6RUN_MEM_physBaseAddr);
}

void *C6RUN_MEM_phys2Virt(void *physPtr)
{
  return (void *) (((physPtr) - C6RUN_MEM_physBaseAddr) + C6RUN_MEM_virtBaseAddr);
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/

static void LOCAL_insertInBufferTable(void *ptr)
{
  uint32_t binNum = GET_BIN_NUMBER(ptr);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];  
  pC6RUN_MEM_Buffer pPtrBuffer = PTR2BUF(ptr);
  
  VERBOSE_PRINT2("Insert - Bin: %d,Pointer: 0x%08X\n",binNum,(uint32_t)ptr);

#if (0)  
  // Order in a queue fashion
  if (pTmpBuffer == NULL)
  {
    C6RUN_MEM_bufferTable[binNum] = pPtrBuffer;
  }
  else
  {
    while (pTmpBuffer->next != NULL)
    {
      pTmpBuffer = pTmpBuffer->next;
      VERBOSE_PRINT1("         Buffer: 0x%08X\n",(uint32_t)pTmpBuffer);
    }
    pTmpBuffer->next = pPtrBuffer;
  }
#else
  // Order in a stack fashion. this should be more beneficial because many times heap allocation
  // occurs with the most recently malloc'ed buffer being the next one to be freed.
  if (pPtrBuffer != NULL)
  {
    pPtrBuffer->next = pTmpBuffer;
  }
  C6RUN_MEM_bufferTable[binNum] = pPtrBuffer;
#endif
}

static void LOCAL_removeFromBufferTable(void *ptr)
{
  uint32_t binNum = GET_BIN_NUMBER(ptr);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];
  pC6RUN_MEM_Buffer pPtrBuffer = PTR2BUF(ptr);
  
  VERBOSE_PRINT2("Remove - Bin: %d,Pointer: 0x%08X\n",binNum,(uint32_t)ptr);

  // We check, but this should never happen
  if (!pTmpBuffer)
  {
    return;
  }

  if (pPtrBuffer == pTmpBuffer)
  {
    C6RUN_MEM_bufferTable[binNum] = pTmpBuffer->next;
    return;
  }
  else
  {
    while (pTmpBuffer->next != NULL)
    {
      if (pPtrBuffer == pTmpBuffer->next)
      {
        pTmpBuffer->next = pTmpBuffer->next->next;
        return;
      }
      pTmpBuffer = pTmpBuffer->next;
      VERBOSE_PRINT1("         Buffer: 0x%08X\n",(uint32_t)pTmpBuffer);
    }
  }
}


/***********************************************************
* End file                                                 *
***********************************************************/
